// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

/*****************************************************************************
   \brief DMA allocation tracking options
*****************************************************************************/

typedef enum AL_ETrackDmaMode
{
  AL_TRACK_DMA_MODE_NONE,
  AL_TRACK_DMA_MODE_SUMMARY,
  AL_TRACK_DMA_MODE_LIVE,
}AL_ETrackDmaMode;
