// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

#include "lib_common_dec/I_DecArch.h"

extern AL_IDecArch decHost;

AL_IDecArch* LibDecodeHostGet(void);
