// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

#include "lib_encode/I_EncArch.h"

extern AL_IEncArch encHost;

AL_IEncArch* LibEncodeHostGet(void);
