// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

#include "lib_decode/DecSettings.h"

/*****************************************************************************/
int32_t GetAlignedStreamBufferSize(int32_t iStreamBufferSize);
