// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

/* This file was generated by configure. DO NOT MODIFY */

#pragma once


#ifndef AL_ENABLE_ENC_WATCHDOG
#define AL_ENABLE_ENC_WATCHDOG 1
#endif
#ifndef AVC_MAX_HORIZONTAL_RANGE_P
#define AVC_MAX_HORIZONTAL_RANGE_P 16
#endif
#ifndef AVC_MAX_VERTICAL_RANGE_P
#define AVC_MAX_VERTICAL_RANGE_P 16
#endif
#ifndef AVC_MAX_HORIZONTAL_RANGE_B
#define AVC_MAX_HORIZONTAL_RANGE_B 8
#endif
#ifndef AVC_MAX_VERTICAL_RANGE_B
#define AVC_MAX_VERTICAL_RANGE_B 8
#endif
#ifndef HEVC_MAX_HORIZONTAL_RANGE_P
#define HEVC_MAX_HORIZONTAL_RANGE_P 32
#endif
#ifndef HEVC_MAX_VERTICAL_RANGE_P
#define HEVC_MAX_VERTICAL_RANGE_P 32
#endif
#ifndef HEVC_MAX_HORIZONTAL_RANGE_B
#define HEVC_MAX_HORIZONTAL_RANGE_B 16
#endif
#ifndef HEVC_MAX_VERTICAL_RANGE_B
#define HEVC_MAX_VERTICAL_RANGE_B 16
#endif
#ifndef ENCODER_DEVICES
#define ENCODER_DEVICES { "/dev/allegroIP" }
#endif
#ifndef ENCODER_CORE_FREQUENCY
#define ENCODER_CORE_FREQUENCY 666666666
#endif
#ifndef ENCODER_CORE_FREQUENCY_MARGIN
#define ENCODER_CORE_FREQUENCY_MARGIN 10
#endif
#ifndef ENCODER_CYCLES_FOR_BLK_32X32
#define ENCODER_CYCLES_FOR_BLK_32X32 { 4900, 4900, 4900, -1 }
#endif
#ifndef DECODER_DEVICES
#define DECODER_DEVICES { "/dev/allegroDecodeIP" }
#endif
#ifndef AL_ENC_NUM_CORES
#define AL_ENC_NUM_CORES 4
#endif
#ifndef AL_ENC_NUM_CORE_JPEG
#define AL_ENC_NUM_CORE_JPEG 0
#endif
#ifndef AL_DEC_NUM_CORES
#define AL_DEC_NUM_CORES 2
#endif
#ifndef AL_DEC_SW_MAX_STACK_SIZE
#define AL_DEC_SW_MAX_STACK_SIZE 16
#endif
#ifndef HW_IP_BURST_ALIGNMENT
#define HW_IP_BURST_ALIGNMENT 32
#endif
#ifndef AL_ENC_CORE_MAX_WIDTH
#define AL_ENC_CORE_MAX_WIDTH 4096
#endif
#ifndef AL_ENC_CORE_MAX_WIDTH_JPEG
#define AL_ENC_CORE_MAX_WIDTH_JPEG 16384
#endif
#ifndef AL_ENC_PITCH_ALIGNMENT
#define AL_ENC_PITCH_ALIGNMENT 32
#endif
#ifndef AL_DEC_PITCH_ALIGNMENT
#define AL_DEC_PITCH_ALIGNMENT 256
#endif
#ifndef AL_MAX_ENC_SLICE
#define AL_MAX_ENC_SLICE 200
#endif
#ifndef AL_LOG_LEVEL
#define AL_LOG_LEVEL 0
#endif
#ifndef MAX_NUM_LAYER
#define MAX_NUM_LAYER 1
#endif
#ifndef AL_TITLE
#define AL_TITLE Xilinx
#endif


#define AL_CONFIGURE_COMMANDLINE "./configure --config-blob customers/Xilinx/config.ini"
#define AL_VERSION_MAJOR 0
#define AL_VERSION_MINOR 48
#define AL_VERSION_PATCH 0
