// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

typedef enum AL_EIDecSchedulerInternalInfo
{
  AL_IDECSCHEDULERINTERNAL_MAX_ENUM,
}AL_EIDecSchedulerInternalInfo;
