// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#include "Itu_Utils.h"

int GetNumActiveRefIdx(AL_TEncChanParam const* pChParam)
{
  (void)pChParam;
  int32_t iNumRef = 1;

  return iNumRef;
}
