// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

int32_t AL_DevicePool_Open(char const* filename);
int32_t AL_DevicePool_Close(int32_t fd);
