// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

/*****************************************************************************
   \brief Select control software architecture
*****************************************************************************/
typedef enum AL_ELibDecoderArch
{
  AL_LIB_DECODER_ARCH_HOST,
}AL_ELibDecoderArch;
