// SPDX-FileCopyrightText: © 2025 Allegro DVT <github-ip@allegrodvt.com>
// SPDX-License-Identifier: MIT

#pragma once

void DisplayFrameStatus(int32_t iFrameNum);
void DisplayVersionInfo();
void DisplayBuildInfo();
